
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '38.5.0',
    'branch_nick': 'release-38-LTS',
    'revision_id': '430fd3dd8f4dc5103e1e932c9515421e1c515c3e',
    'revno': 14073
}
